"""
Visualize content of a Bluesky run.

* BRC: BlueskyRunsCatalog

.. autosummary::

    ~BRCRunVisualization
"""

from PyQt5 import QtWidgets

from . import utils


class BRCRunVisualization(QtWidgets.QWidget):
    """The panel to show the contents of a run."""

    # UI file name matches this module, different extension
    ui_file = utils.getUiFileName(__file__)

    def __init__(self, parent):
        self.parent = parent

        super().__init__()
        utils.myLoadUi(self.ui_file, baseinstance=self)
        self.setup()

    def setup(self):
        pass

    def setMetadata(self, text, *args, **kwargs):
        self.metadata.setText(text)

    def setData(self, text, *args, **kwargs):
        self.data.setText(text)

    def setPlot(self, plot_widget):
        layout = self.plotPage.layout()
        utils.removeAllLayoutWidgets(layout)
        layout.addWidget(plot_widget)
        self.tabWidget.setCurrentWidget(self.plotPage)

    def setStatus(self, text):
        self.parent.setStatus(text)


# -----------------------------------------------------------------------------
# :copyright: (c) 2023-2025, UChicago Argonne, LLC
#
# Distributed under the terms of the Argonne National Laboratory Open Source License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------



"""
TAPI: Local support for the tiled API & data structures.

.. autosummary:

    ~connect_tiled_server
    ~get_tiled_runs
    ~QueryTimeSince
    ~QueryTimeUntil
    ~RunMetadata
    ~TiledServerError
"""

import logging

import tiled
import tiled.queries
from httpx import HTTPStatusError

logger = logging.getLogger(__name__)


class TiledServerError(RuntimeError):
    """An error from the tiled server."""


class RunMetadata:
    """Cache the metadata for a single run."""

    def __init__(self, cat, uid):
        self.catalog = cat
        self.uid = uid
        self.request_from_tiled_server()

    def __str__(self) -> str:
        return (
            f"{__class__.__name__}(catalog={self.catalog.item['id']!r},"
            f" uid7={self.uid[:7]!r},"
            f" active={self.active})"
        )

    def request_from_tiled_server(self):
        """Get run details from server."""
        self.run = self.catalog[self.uid]
        self.run_md = self.run.metadata
        self.active = (
            self.uid == self.catalog.keys().last() and "stop" not in self.run_md
        )
        self.streams_md = None
        self.streams_data = None

    def get_run_md(self, doc, key, default=None):
        """Get metadata by key from run document."""
        return (self.run_md.get(doc) or {}).get(key, default)

    def plottable_signals(self):
        """
        Return a dict with the plottable data for this run.

        * field: any available numeric data keys
        * motors: any data keys for motors declared by the run
        * detectors: any numeric data keys that are not motors or excluded names
        * plot_signal: the first detector signal
        * plot_axes: the first motor signal for each dimension

        * run.metadata[hints][dimensions] show the independent axes object names
        * Any given dimension may have more than one motor object (a2scan, ...)
        * This code chooses to only use the first motor of each dimension.
        * The stream descriptor list is usually length = 1.
        * object_keys are used to get lists of data_keys (fields)
        """

        def find_name_device_or_signal(key):
            if key in stream_hints:  # from ophyd.Device
                return stream_hints[key]["fields"]
            elif key in descriptor["data_keys"]:  # from ophyd.Signal
                return [key]
            raise KeyError(f"Could not find {key=}")

        def get_signal(key):
            try:
                return find_name_device_or_signal(key)[0]  # just the first one
            except KeyError:
                if key != "time":
                    raise KeyError(f"Unexpected key: {key!r}")
                return key  # "time" is a special case

        def is_numeric(signal):
            dtype = descriptor["data_keys"][signal]["dtype"]
            if dtype == "array":
                stream_data = self.stream_data(self.stream_name)
                ntype = stream_data["data"][signal].dtype.name
                if ntype.startswith("int") or ntype.startswith("float"):
                    dtype = "number"
            return dtype == "number"

        # dimensions of the run
        run_dims = self.get_run_md("start", "hints", {}).get("dimensions", [])

        # data stream to be used
        streams = [d[1] for d in run_dims]
        if len(set(streams)) != 1:
            # All hinted dimensions should come from the same stream.
            raise ValueError(f"Not handling hinted dimensions: {run_dims=}")
        stream = streams[0]

        # description of the data stream objects
        descriptors = self.stream_metadata(stream).get("descriptors", {})
        if len(descriptors) != 1:
            raise ValueError(f"Not handling situation of {len(descriptors)=}")

        descriptor = descriptors[0]

        # Mapping from object_keys to data_keys.
        stream_hints = descriptor.get("hints", {})

        # First motor signal for each dimension.
        try:
            axes = [get_signal(d[0][0]) for d in run_dims]
        except KeyError as exc:
            raise exc

        # All motor signals.
        motors = [
            signal
            for motor in self.get_run_md("start", "motors", [])
            for signal in find_name_device_or_signal(motor)
        ]

        # All detector signals.
        detectors = [
            signal
            for detector in self.get_run_md("start", "detectors")
            for signal in find_name_device_or_signal(detector)
            if is_numeric(signal)
        ]

        fields = []
        for descriptor in descriptors:
            hints = descriptor.get("hints", {})
            for obj_name in descriptor["object_keys"]:
                try:
                    signals = hints.get(obj_name, {})["fields"]
                except KeyError:
                    # ``ophyd.Device`` can have multiple signals
                    signals = descriptor["object_keys"].get(obj_name, [])
                fields.extend([k for k in signals if is_numeric(k)])

        status = self.get_run_md("stop", "exit_status")
        plot_signal = None
        if status in "abort success".split():
            # These runs probably have plottable data fields.

            # Do not choose any of these fields as the default
            # (NeXus-style plottable) signal data.
            not_plottable_signals = """
                timebase
                preset_time
            """.split()

            names_to_avoid = motors + not_plottable_signals
            possible_signals = detectors + fields
            for field in possible_signals:
                if field not in names_to_avoid:
                    plot_signal = field
                    break

        return {
            "catalog": self.catalog.item["id"],
            "uid": self.uid,
            "stream": stream,
            "plot_signal": plot_signal,
            "plot_axes": axes,
            "motors": motors,
            "detectors": detectors,
            "fields": fields,
        }

    def stream_data(self, stream_name):
        """Return the data structure for this stream."""
        if self.streams_data is None:
            # Optimize with a cache.
            self.streams_data = {
                sname: self.run[sname]["data"].read() for sname in self.run
            }

        return self.streams_data[stream_name]

    def stream_data_field_shape(self, stream_name, field_name):
        """Shape of this data field."""
        stream = self.stream_data(stream_name)
        try:
            shape = stream[field_name].shape
        except Exception:
            shape = ()
        return shape

    def stream_data_fields(self, stream_name):
        """
        Data field (names) of this BlueskyEventStream.

        Sort the list by relevance.

        First "time" (epoch timestamp for each event document), then "config" (the
        caller provided these names as parameters for this stream), then "data"
        (other signals in this stream, usually added from a Device hint).
        """
        fields = sorted(self.stream_data(stream_name))

        # Promote "time" field to first place.
        if "time" in fields:
            fields.remove("time")
        fields.insert(0, "time")
        return fields

    def stream_data_field_pv(self, stream_name, field_name):
        """EPICS PV name of this field."""
        pv = ""
        try:
            descriptors = self.stream_metadata(stream_name).get("descriptors", {})
            assert len(descriptors) == 1, f"{stream_name=} has {len(descriptors)=}"
            source = descriptors[0]["data_keys"][field_name].get("source", "")
            if source.startswith("PV:"):
                pv = source[3:]
        except Exception:
            pass
        return pv

    def stream_data_field_units(self, stream_name, field_name):
        """Engineering units of this field."""
        units = ""
        try:
            descriptors = self.stream_metadata(stream_name).get("descriptors", {})
            assert len(descriptors) == 1, f"{stream_name=} has {len(descriptors)=}"
            units = descriptors[0]["data_keys"][field_name].get("units", "")
        except Exception:
            pass
        return units

    def stream_metadata(self, stream_name=None):
        """Return the metadata dictionary for this stream."""
        if self.streams_md is None:
            # Optimize with a cache.
            self.streams_md = {sname: self.run[sname].metadata for sname in self.run}

        if stream_name is None:
            return self.streams_md
        return self.streams_md[stream_name]

    def summary(self):
        """Summary (text) of this run."""
        return (
            f"{self.get_run_md('start', 'scan_id', '?')}"
            f" {self.get_run_md('start', 'plan_name', '')}"
            f" {self.get_run_md('start', 'title', '')}"
        ).strip()


def connect_tiled_server(uri):
    """Make connection with the tiled server URI.  Return a client object."""
    from tiled.client import from_uri

    # leave out "dask" and get numpy by default
    # https://examples.dask.org/array.html
    # > Call .compute() when you want your result as a NumPy array.
    client = from_uri(uri, "dask")
    return client


def get_tiled_slice(cat, offset, size, ascending=True):
    end = offset + size
    key_gen = cat.keys()

    try:
        return key_gen[offset:end]
    except HTTPStatusError as exc:
        # fmt: off
        # logger.error("HTTPStatusError: %s", exc)
        raise TiledServerError(
            f"{exc.response.reason_phrase}"
            f" ({exc.response.status_code})"
            "  Adjust filters to reduce the catalog size."
        )
        # fmt: on


def QueryTimeSince(isotime):
    """Tiled client query: all runs since given date/time."""
    from . import utils

    return tiled.queries.Key("time") >= utils.iso2ts(isotime)


def QueryTimeUntil(isotime):
    """Tiled client query: all runs until given date/time."""
    from . import utils

    return tiled.queries.Key("time") <= utils.iso2ts(isotime)


def get_run(uri=None, catalog="training", reference=None):
    """Get referenced run from tiled server catalog."""
    # from gemviz.tapi import connect_tiled_server
    # from gemviz.tapi import get_tiled_runs

    uri = uri or "http://localhost:8020"
    client = connect_tiled_server(uri)
    cat = get_tiled_runs(client[catalog], plan_name="scan")
    reference = reference or -1
    uid = reference if isinstance(reference, str) else cat.keys()[reference]
    run = cat[uid]
    return run


def get_tiled_runs(cat, since=None, until=None, text=[], text_case=[], **keys):
    """
    Return a new catalog, filtered by search terms.

    Runs will be selected with start time `>=since` and `< until`.
    If either is `None`, then the corresponding filter will not be
    applied.

    Parameters

    `cat` obj :
        This is the catalog to be searched.
        `Node` object returned by tiled.client.
    `since` str :
        Earliest start date (& time), in ISO8601 format.
    `until` str :
        Latest start date (& time), in ISO8601 format.
    `text` [str] :
        List of full text searches.  Not sensitive to case.
    `text_case` [str] :
        List of full text searches.  Case sensitive.
    `keys` dict :
        Dictionary of metadata keys and values to be matched.
    """
    if since is not None:
        cat = cat.search(QueryTimeSince(since))
    if until is not None:
        cat = cat.search(QueryTimeUntil(until))

    for k, v in keys.items():
        cat = cat.search(tiled.queries.Key(k) == v)

    for v in text:
        cat = cat.search(tiled.queries.FullText(v, case_sensitive=False))
    for v in text_case:
        cat = cat.search(tiled.queries.FullText(v, case_sensitive=True))
    return cat


# -----------------------------------------------------------------------------
# :copyright: (c) 2023-2025, UChicago Argonne, LLC
#
# Distributed under the terms of the Argonne National Laboratory Open Source License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------



"""
QWidget to select stream data fields for plotting.

.. autosummary::

    ~SelectFieldsWidget
    ~to_datasets
"""

import datetime
import logging

import xarray
from PyQt5 import QtCore
from PyQt5 import QtWidgets

from . import utils
from .select_fields_tablemodel import ColumnDataType
from .select_fields_tablemodel import FieldRuleType
from .select_fields_tablemodel import TableColumn
from .select_fields_tablemodel import TableField
from .select_fields_tableview import SelectFieldsTableView

logger = logging.getLogger(__name__)
DEFAULT_STREAM = "primary"

STREAM_COLUMNS = [
    TableColumn("Field", ColumnDataType.text),
    TableColumn("X", ColumnDataType.checkbox, rule=FieldRuleType.unique),
    TableColumn("Y", ColumnDataType.checkbox, rule=FieldRuleType.multiple),
    TableColumn("Shape", ColumnDataType.text),
]


class SelectFieldsWidget(QtWidgets.QWidget):
    """Panel to select fields (signals) for plotting."""

    ui_file = utils.getUiFileName(__file__)
    selected = QtCore.pyqtSignal(str, str, dict)

    def __init__(self, parent, run):
        self.parent = parent
        self.run = run  # tapi.RunMetadata object
        self.analysis = run.plottable_signals()
        self.stream_name = self.analysis.get("stream", DEFAULT_STREAM)

        super().__init__()
        utils.myLoadUi(self.ui_file, baseinstance=self)
        self.setup()

    def setup(self):
        self.run_summary.setText(self.run.summary())

        stream_list = list(self.run.stream_metadata())
        if "baseline" in stream_list:
            # Too many signals! 2 points each.  Do not plot from "baseline" stream.
            stream_list.pop(stream_list.index("baseline"))
        index = stream_list.index(self.stream_name)
        if index > 0:
            # Move the default stream to the first position.
            stream_list.insert(0, stream_list.pop(index))

        if len(stream_list) > 0:
            self.setStream(stream_list[0])

            self.streams.clear()
            self.streams.addItems(stream_list)
            self.streams.currentTextChanged.connect(self.setStream)

    def setStream(self, stream_name):
        from functools import partial

        self.stream_name = stream_name
        stream = self.run.run[stream_name]
        logger.debug("stream_name=%s, stream=%s", stream_name, stream)

        x_names = self.analysis["plot_axes"]
        y_name = self.analysis["plot_signal"]

        # describe the data fields for the dialog.
        sdf = self.run.stream_data_fields(stream_name)
        # print(f"{__name__}.{__class__.__name__}: {sdf=}")
        fields = []
        for field_name in sdf:
            selection = None
            if x_names is not None and field_name in x_names:
                selection = "X"
            elif y_name is not None and field_name == y_name:
                selection = "Y"
            shape = self.run.stream_data_field_shape(stream_name, field_name)
            if len(shape) == 0:
                # print(f"{stream_name=} {field_name=} {shape=}")
                logger.debug(
                    "stream_name=%s field_name=%s shape=%s",
                    stream_name,
                    field_name,
                    shape,
                )
            field = TableField(field_name, selection=selection, shape=shape)
            fields.append(field)
        logger.debug("fields=%s", fields)

        # build the view of this stream
        view = SelectFieldsTableView(self)
        view.displayTable(STREAM_COLUMNS, fields)
        view.selected.connect(partial(self.relayPlotSelections, stream_name))

        layout = self.groupbox.layout()
        utils.removeAllLayoutWidgets(layout)
        layout.addWidget(view)

    def relayPlotSelections(self, stream_name, action, selections):
        """Receive selections from the dialog and relay to the caller."""
        # selections["stream_name"] = self.stream_name
        self.selected.emit(stream_name, action, selections)


def to_datasets(run, stream_name, selections, scan_id=None):
    """Prepare datasets and options for plotting."""
    from . import chartview

    stream = run.stream_data(stream_name)

    x_axis = selections.get("X")
    x_datetime = False  # special scaling using datetime
    if x_axis is None:
        x_data = None
        x_units = ""
        x_axis = "data point number"
    else:
        x_data = stream[x_axis].compute()
        x_shape = x_data.shape
        x_units = run.stream_data_field_units(stream_name, x_axis)
        if len(x_shape) != 1:
            # fmt: off
            raise ValueError(
                "Can only plot 1-D data now."
                f" {x_axis} shape is {x_shape}"
            )
            # fmt: on
        if x_axis == "time" and min(x_data) > chartview.TIMESTAMP_LIMIT:
            x_units = ""
            x_datetime = True
            x_data = xarray.DataArray(
                data=list(map(datetime.datetime.fromtimestamp, x_data[x_axis].data)),
                name=x_axis,
                # dims=x_axis,
                # coords=?,
            )

    datasets = []
    y_selections = selections.get("Y", [])
    if len(y_selections) == 0:
        raise ValueError("No Y data selected.")
    for y_axis in y_selections:
        ds, ds_options = [], {}
        color = chartview.auto_color()
        symbol = chartview.auto_symbol()

        y_data = stream[y_axis].compute()
        y_units = run.stream_data_field_units(stream_name, y_axis)
        y_shape = y_data.shape
        if len(y_shape) != 1:
            # fmt: off
            raise ValueError(
                "Can only plot 1-D data now."
                f" {y_axis} shape is {y_shape}"
            )

        # keys used here must match the plotting back-end (matplotlib)
        scan_id = run.get_run_md("start", "scan_id")
        # verbose labels
        # ds_options["label"] = f"{y_axis} ({run.summary()} {run.uid[:7]})"
        # terse labels
        ds_options["label"] = f"{scan_id} ({run.uid[:7]})"
        ds_options["color"] = color  # line color
        ds_options["marker"] = symbol
        ds_options["markersize"] = 5  # default: 10

        if x_data is None:
            ds = [y_data]  # , title=f"{y_axis} v index"
        else:
            if x_shape != y_shape:
                raise ValueError(
                    "Cannot plot.  Different shapes for"
                    f" X ({x_shape!r})"
                    f" and Y ({y_shape!r}) data."
                )
            ds = [x_data, y_data]
        datasets.append((ds, ds_options))

    plot_options = {
        "x_datetime": x_datetime,
        "x_units": x_units,
        "x": x_axis,
        "y_units": y_units,
        "y": ", ".join(selections.get("Y", [])),
    }

    return datasets, plot_options


# -----------------------------------------------------------------------------
# :copyright: (c) 2023-2025, UChicago Argonne, LLC
#
# Distributed under the terms of the Argonne National Laboratory Open Source License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------



"""
Qt widget that shows the plot.

.. autosummary::

    ~auto_color
    ~auto_symbol
    ~ChartView
    ~PLOT_COLORS
    ~PLOT_SYMBOLS
    ~TIMESTAMP_LIMIT

:see: https://matplotlib.org/stable/users/index.html

..  note:: To see the full list of plot symbols from
    MatPlotLib (https://matplotlib.org/stable/gallery/lines_bars_and_markers/marker_reference.html)::

        from matplotlib.lines import Line2D
        print(Line2D.markers)

"""

import datetime
from itertools import cycle

from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.backends.backend_qt5agg import NavigationToolbar2QT as NavigationToolbar
from matplotlib.figure import Figure
from PyQt5 import QtWidgets

TIMESTAMP_LIMIT = datetime.datetime.fromisoformat("1990-01-01").timestamp()
"""Earliest date for a run in any Bluesky catalog (1990-01-01)."""

PLOT_COLORS = """
    r g b c m
    goldenrod
    lime
    orange
    blueviolet
    brown
    teal
    olive
    lightcoral
    gold
    cornflowerblue
    forestgreen
    salmon
""".split()
"""
Select subset of the MatPlotLib named colors.

Do **NOT** sort these colors alphabetically!  There should be obvious
contrast between adjacent colors.

* :see: https://matplotlib.org/stable/gallery/color/named_colors.html
* :see: https://developer.mozilla.org/en-US/docs/Web/CSS/named-color
"""

PLOT_SYMBOLS = """o + x * s d ^ v""".split()
"""
Select subset of the MatPlotLib marker symbols.

To print the full dictionary of symbols available::

    from matplotlib.lines import Line2D
    print(Line2D.markers)

:see: https://matplotlib.org/stable/gallery/lines_bars_and_markers/marker_reference.html
"""

# iterators for colors & symbols
_AUTO_COLOR_CYCLE = cycle(PLOT_COLORS)
_AUTO_SYMBOL_CYCLE = cycle(PLOT_SYMBOLS)


def auto_color():
    """Returns next color for pens and brushes."""
    return next(_AUTO_COLOR_CYCLE)


def auto_symbol():
    """Returns next symbol for scatter plots."""
    return next(_AUTO_SYMBOL_CYCLE)


class ChartView(QtWidgets.QWidget):
    """
    MatPlotLib Figure

    .. autosummary::

        ~plot
        ~setAxisLabel
        ~setAxisUnits
        ~setBottomAxisText
        ~setLeftAxisText
        ~setLeftAxisUnits
        ~setPlotTitle
    """

    def __init__(self, parent, **kwargs):
        self.parent = parent

        super().__init__()

        size = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred
        )

        # Remember these Matplotlib figure, canvas, and axes objects.
        self.figure = Figure()
        self.canvas = FigureCanvas(self.figure)
        self.main_axes = self.figure.add_subplot(111)

        # Adjust margins
        self.figure.subplots_adjust(bottom=0.1, top=0.9, right=0.92)
        self.setOptions()

        config = {
            "title": self.setPlotTitle,
            "subtitle": self.setPlotSubtitle,
            "y": self.setLeftAxisText,
            "x": self.setBottomAxisText,
            "x_units": self.setBottomAxisUnits,
            "y_units": self.setLeftAxisUnits,
            "x_datetime": self.setAxisDateTime,
        }
        for k, func in config.items():
            func(kwargs.get(k))

        # QWidget Layout
        layout = QtWidgets.QVBoxLayout()
        self.setLayout(layout)
        # Add directly unless we plan to use the toolbar later.
        layout.addWidget(NavigationToolbar(self.canvas, self))
        layout.addWidget(self.canvas)

        # plot
        size.setHorizontalStretch(4)

        self.curves = {}  # all the curves on the graph, key = label

    def addCurve(self, *args, title="plot title", **kwargs):
        """Add to graph."""
        plot_obj = self.main_axes.plot(*args, **kwargs)
        self.updatePlot(title)
        # Add to the dictionary
        label = kwargs.get("label")
        if label is None:
            raise KeyError("This curve has no label.")
        self.curves[label] = plot_obj[0], *args

    def option(self, key, default=None):
        return self.plotOptions().get(key, default)

    def plot(self, *args, title="plot title", **kwargs):
        """
        Plot from the supplied (x, y) or (y) data.

        PARAMETERS

        - args tuple: x & y xarray.DataArrays.  When only y is supplied, x will
          be the index.
        - kwargs (dict): dict(str, obj)
        """
        self.setOptions(**kwargs.get("plot_options", {}))
        ds_options = kwargs.get("ds_options", kwargs)
        self.main_axes.axis("on")

        label = ds_options.get("label")
        if label is None:
            raise KeyError("This curve has no label.")
        if label not in self.curves:
            self.addCurve(*args, title=title, **ds_options)

    def plotOptions(self):
        return self._plot_options

    def setAxisDateTime(self, choice):
        pass  # data provided in datetime objects

    def setAxisLabel(self, axis, text):
        set_axis_label_method = {
            "bottom": self.main_axes.set_xlabel,
            "left": self.main_axes.set_ylabel,
        }[axis]
        set_axis_label_method(text)

    def setAxisUnits(self, axis, text):
        pass  # TODO: not implemented yet

    def setBottomAxisText(self, text):
        self.setAxisLabel("bottom", text)

    def setBottomAxisUnits(self, text):
        self.setAxisUnits("bottom", text)

    def setConfigPlot(self, grid=True):
        self.setLeftAxisText(self.ylabel())
        self.setBottomAxisText(self.xlabel())
        self.setPlotTitle(self.title())
        if grid:
            self.main_axes.grid(True, color="#cccccc", linestyle="-", linewidth=0.5)
        else:
            self.main_axes.grid(False)
        self.canvas.draw()

    def setLeftAxisText(self, text):
        self.setAxisLabel("left", text)

    def setLeftAxisUnits(self, text):
        self.setAxisUnits("left", text)

    def setOption(self, key, value):
        self._plot_options[key] = value

    def setOptions(self, **options):
        self._plot_options = options

    def setPlotTitle(self, text):
        if text is not None:
            self.figure.suptitle(text)

    def setPlotSubtitle(self, text):
        if text is not None:
            self.main_axes.set_title(text, size=7, x=1, ha="right", color="lightgrey")

    def setSubtitle(self, text):
        self.setOption("subtitle", text)

    def setTitle(self, text):
        self.setOption("title", text)

    def setXLabel(self, text):
        self.setOption("xlabel", text)

    def setYLabel(self, text):
        self.setOption("ylabel", text)

    def subtitle(self):
        return self.option("subtitle")

    def title(self):
        return self.option("title")

    def updateLegend(self):
        labels = self.main_axes.get_legend_handles_labels()[1]
        valid_labels = [label for label in labels if not label.startswith("_")]
        if valid_labels:
            self.main_axes.legend()

    def updatePlot(self, title):
        """Update annotations (titles & axis labels)."""
        self.setPlotTitle(title)

        iso8601 = datetime.datetime.now().isoformat(sep=" ", timespec="seconds")
        subtitle = f"plotted: {iso8601}"
        if self.parent is not None:
            cat_name = self.parent.catalogName() or ""
            subtitle = f"catalog={cat_name!r}  {subtitle}"
        self.setPlotSubtitle(subtitle)

        self.setBottomAxisText(self.xlabel())
        self.setLeftAxisText(self.ylabel())

        # Recompute the axes limits and autoscale:
        self.main_axes.relim()
        self.main_axes.autoscale_view()
        self.updateLegend()
        self.setConfigPlot()
        self.canvas.draw()

    def xlabel(self):
        return self.option("xlabel")

    def ylabel(self):
        return self.option("ylabel")


# -----------------------------------------------------------------------------
# :copyright: (c) 2023-2025, UChicago Argonne, LLC
#
# Distributed under the terms of the Argonne National Laboratory Open Source License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------



"""
QTableView of tiled "CatalogOfBlueskyRuns".

BRC: BlueskyRunsCatalog

Uses :class:`bluesky_runs_catalog_table_model.BRCTableModel`.

.. autosummary::

    ~BRCTableView
"""

import logging
from functools import partial

from PyQt5 import QtCore
from PyQt5 import QtWidgets

from . import utils

logger = logging.getLogger(__name__)


class BRCTableView(QtWidgets.QWidget):
    ui_file = utils.getUiFileName(__file__)
    run_selected = QtCore.pyqtSignal(object)

    def __init__(self, parent, catalog, page_offset, page_size):
        self.parent = parent
        self._catalog = catalog
        self._catalog_length = len(catalog)
        self.run_cache = {}

        super().__init__(parent)
        utils.myLoadUi(self.ui_file, baseinstance=self)
        self.setup(page_offset, page_size)

    def setup(self, page_offset, page_size):
        """Setup the catalog view panel."""
        from .bluesky_runs_catalog_table_model import BRCTableModel

        self.model = BRCTableModel(self)
        self.tableView.setModel(self.model)

        # since we cannot set header's ResizeMode in Designer ...
        header = self.tableView.horizontalHeader()
        header.setSectionResizeMode(QtWidgets.QHeaderView.ResizeToContents)

        if self.pageSize.findText(str(page_size)) == -1:
            self.pageSize.insertItem(0, str(page_size))
        self.pageSize.setCurrentText(str(page_size))
        self.setPage(page_offset, page_size)

        self.pageSize.currentTextChanged.connect(
            partial(self.doPagerButtons, "pageSize")
        )
        for button_name in "first back next last".split():
            button = getattr(self, button_name)
            # custom: pass the button name to the receiver
            button.released.connect(partial(self.doPagerButtons, button_name))

        self.parent.brc_search_panel.enableDateRange(
            len(self.parent.brc_search_panel.catalog()) > 0
        )

        self.setButtonPermissions()
        self.setPagerStatus()
        self.tableView.clicked.connect(self.doRunSelectedSlot)

    def doPagerButtons(self, action, **kwargs):
        """User clicked a button to change the page."""
        logger.debug("action=%s", action)

        if action == "first":
            self.setPage(0, self.page_size)
        elif action == "back":
            self.setPage(self.page_offset - self.page_size, self.page_size)
        elif action == "pageSize":
            self.setPage(self.page_offset, self.pageSize.currentText())
        elif action == "next":
            self.setPage(self.page_offset + self.page_size, self.page_size)
        elif action == "last":
            self.setPage(-1, self.page_size)

        self.setButtonPermissions()
        self.setPagerStatus()

    @property
    def pagerAtStart(self):
        """Is this the first page?"""
        return self.page_offset == 0

    @property
    def pagerAtEnd(self):
        """Is this the last page?"""
        # number is zero-based
        return (self.page_offset + self.page_size) >= self.catalogLength()

    def setButtonPermissions(self):
        """Enable/disable the pager buttons, depending on page in view."""
        first_page = self.pagerAtStart
        last_page = self.pagerAtEnd

        self.first.setEnabled(not first_page)
        self.back.setEnabled(not first_page)
        self.next.setEnabled(not last_page)
        self.last.setEnabled(not last_page)

    def setPage(self, offset, size):
        """Choose the page.  Update the model."""
        # user cannot edit directly, not expected to raise an exception
        offset = int(offset)
        size = int(size)

        self.page_size = max(0, min(size, self.catalogLength()))
        if offset >= 0:
            offset = min(offset, self.catalogLength() - self.page_size)
        else:
            offset = self.catalogLength() - self.page_size
        self.page_offset = max(0, offset)
        if int(self.pageSize.currentText()) != self.page_size:
            self.pageSize.setCurrentText(str(self.page_size))
        logger.debug(
            "len(catalog)=%d  offset=%d  size=%d",
            self.catalogLength(),
            self.page_offset,
            self.page_size,
        )

        # TODO: unselect row if selection is not on the page
        # see: https://stackoverflow.com/questions/64225673
        # "how-to-deselect-an-entire-qtablewidget-row"

        self.updateModelData()

    def updateModelData(self):
        """Send a new page of runs to the model."""
        from . import tapi

        # get list of metadata for each run to be shown in the table
        start = self.page_offset
        end = self.page_offset + self.page_size
        uid_list = self.catalog().keys()[start:end]

        page = {}  # the new page of run metadata
        for uid in uid_list:
            run_md = self.run_cache.get(uid)
            if run_md is None or run_md.active:
                # Get new information from the server about this run.
                run_md = tapi.RunMetadata(self.catalog(), uid)
                self.run_cache[uid] = run_md  # update the cache
            page[uid] = run_md

        # Send the page of runs to the model now.
        self.model.setRuns(page)

    def setPagerStatus(self, text=None):
        if text is None:
            total = self.catalogLength()  # filtered catalog
            if total == 0:
                text = "No runs"
            else:
                start = self.page_offset
                end = start + self.page_size
                text = f"{start + 1}-{end} of {total} runs"

        self.status.setText(text)
        self.setStatus(text)

    def doRunSelectedSlot(self, index):
        run_md = list(self.model.runs.values())[index.row()]
        self.run_selected.emit(run_md)

    def setCatalog(self, catalog):
        self._catalog = catalog  # filtered catalog
        self._catalog_length = len(catalog)

        uid = self.parent.selected_run_uid
        if uid in self.model.runs:
            offset = list(self.model.runs.keys()).index(uid)
        else:
            offset = -1
        self.setPage(offset, self.page_size)  # ... and update the model
        self.setPagerStatus()

    def catalog(self):
        return self._catalog

    def catalogLength(self):
        # Avoid a bug in the tiled client.  When the client is asked for
        # len(catalog) frequently, it will begin to return a length of ``1``
        # instead of the actual length.  After waiting a short while, the client
        # will return the actual length again.
        return self._catalog_length

    def setStatus(self, text):
        self.parent.setStatus(text)


# -----------------------------------------------------------------------------
# :copyright: (c) 2023-2025, UChicago Argonne, LLC
#
# Distributed under the terms of the Argonne National Laboratory Open Source License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
